
/*	
	This file is part of miSCellaneous, a program library for SuperCollider 3
	
	Created: 2010-10-21, version 0.3
	Copyright (C) 2009, 2010 Daniel Mayer 
	Email: 	daniel-mayer@email.de
	URL:		www.daniel-mayer.at

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


VarGui	{
	var <>varCtr, <>synthCtr, <>synthIDs, <>window,
		<>colors, <>server, <>players, <>varCtrIDs, <>synthCtrIDs, <>varCtrNum, <>synthCtrNum,
		<>varStrings, <>varCtrSliders, <>synthCtrStrings, <>synthCtrSliders, <>synthNameBoxes, <>varCtrColors, <>synthCtrColors,
		<>updateButton, <>stopButton, <>saveButton, <>saveData, <>synthNameData, <>environment;
	
	*new { |varCtr, synthCtr, synthIDs, varCtrGroups, synthCtrGroups, environment, players, server|
		^super.new.init(varCtr, synthCtr, synthIDs, varCtrGroups, synthCtrGroups, environment, players, server);
	}
	*load { |pathname, filename, synthIDs, varCtrGroups, synthCtrGroups, environment, players, server|
		var loadData = (pathname ++ filename).load;
		^super.new.init(loadData[0], loadData[1], synthIDs, varCtrGroups, synthCtrGroups, environment, players, server);
	}
	init { |varCtr, synthCtr, synthIDs, varCtrGroups, synthCtrGroups, environment, players, server|
		var k = 0;
		this.varCtr = varCtr;
		this.synthCtr = synthCtr;
		this.synthIDs = synthIDs;
		this.environment = environment ?? currentEnvironment;
		this.server = server ?? Server.local;
		this.players = players;
		
		varCtrIDs = Dictionary.new;
		saveData = [varCtr.deepCopy, synthCtr.deepCopy]; 
		varCtr.do({ |item,i|
			(i % 2 == 0).if {
				varCtr[i+1].every(_.isSequenceableCollection).if {
					varCtr[i+1].do({|x,l|
						varCtrIDs = varCtrIDs.add(k -> [item,l]);
						k = k + 1;
					})
				}{
					varCtrIDs = varCtrIDs.add(k -> item);
					k = k + 1;
				};
			}
		});
		varCtrNum = varCtrIDs.size;
		varCtrColors = varCtrNum.groupIndices(this.initVarCtrGroups(varCtrGroups));
		k = 0;
		
		synthCtrIDs = Dictionary.new;
		synthCtr.do({ |item1,i|
			i.even.if {
				synthCtr[i+1].do({ |item2,j|
					j.even.if {
						synthCtr[i+1][j+1].every(_.isSequenceableCollection).if {
							synthCtr[i+1][j+1].do({|x,l|
								synthCtrIDs = synthCtrIDs.add(k -> [synthCtr[i], [item2, l]]);
								k = k + 1;
							})
						}{
							synthCtrIDs = synthCtrIDs.add(k -> [synthCtr[i], item2]);
							k = k + 1;
						};
					}
				})
			}
		});
		synthCtrNum = synthCtrIDs.size;
		synthCtrColors = synthCtrNum.groupIndices(this.initSynthCtrGroups(synthCtrGroups));
	}
	
	
	
	gui { |sliderHeight = 24, font = "Arial", colorsLo = 0.25, colorsHi = 0.7, fontHeight, buttonHeightFactor = 1.2, buttonFontFactor = 1|
		var w, wName, colNum = varCtrNum.sign + synthCtrNum.sign, tempColors, totalHeight, totalWidth, heightK,
			minFontHeight = 10, maxFontHeight = 16, minFontSliderHeight = 12, maxFontSliderHeight = 24, 
			cview, varColorNum, ctrColorNum, colorNum, sliderWidth = 330, labelWidth = 80, colMargin = 30, leftMargin = 50, secondColumnIndent,
			rightMargin = 50, upperMargin = 30, lowerMargin = 100, gap = 4, buttonRow, buttonIndex, buttonColorLo = 0.3, buttonColorHi = 0.9,
			xPosMin = 100, xPosMax = 500, yPosMin = 100, yPosMax = 500, buttonGap, buttonRowShift,
			colorAreaUsage = 0.25, colorSteps = 1000, numColorExp = 0.4, k;
		
		varColorNum = (this.varCtrColors.maxItem ?? - 1) + 1;
		ctrColorNum = (this.synthCtrColors.maxItem ?? - 1) + 1;
		colorNum = varColorNum + ctrColorNum + 2;
		varStrings = []; varCtrSliders = []; synthCtrStrings = []; synthCtrSliders = []; synthNameBoxes = [];
		totalHeight = (max(varCtrNum, synthCtrNum) + 2) * (sliderHeight + gap) + (buttonHeightFactor * sliderHeight) + lowerMargin;
		totalWidth = leftMargin + rightMargin +
			(varCtrNum.sign * sliderWidth) +
			(synthCtrNum.sign * (sliderWidth + labelWidth)) +
			((colNum == 2).if { colMargin } { 0 }) + gap;
		tempColors = colorNum.distinctColors(redLo: colorsLo, redHi: colorsHi,
			blueLo: colorsLo, blueHi: colorsHi,
			greenLo: colorsLo, greenHi: colorsHi, 
			areaUsage: colorAreaUsage);
		buttonGap = (varCtrNum == 0).if {
				sliderWidth + gap - (2 * labelWidth) / 2
			}{
				sliderWidth - (3 * labelWidth) / 2
			};

		wName = case 
			{ (varCtrNum != 0) && (synthCtrNum != 0) } { "variable and synth control" }
			{ (varCtrNum == 0) && (synthCtrNum != 0) } { "synth control" }
			{ (varCtrNum != 0) && (synthCtrNum == 0) } { "variable control" }
			{ (varCtrNum == 0) && (synthCtrNum != 0) } { "there is no control" };
		
		w = GUI.window.new(wName, Rect(rrand(xPosMin, xPosMax), rrand(yPosMin, yPosMax), totalWidth, totalHeight));
		cview = GUI.compositeView.new(w, Rect(0, 0, totalWidth, totalHeight));
		cview.decorator = FlowLayout(cview.bounds, leftMargin @ upperMargin);
		cview.resize_(5);
		cview.background = Gradient(tempColors[colorNum - 1], tempColors[colorNum - 2], \v, colorSteps);
		k = 0;
		forBy(0, varCtr.size - 2, 2, {|i|
			var str, ez;
    			varCtr[i+1].every(_.isSequenceableCollection).if {
    				varCtr[i+1].do({|item,j|
   			 		str = varCtr[i].asString ++ "[" ++ j.asString ++ "]" ++ " ";
   			 		cview.decorator.shift(80);
    		 			cview.decorator.nextLine;
		    			varStrings = varStrings.add(str);
    					varCtrSliders = varCtrSliders.add(
    						ez = EZSlider(cview, sliderWidth @ sliderHeight, str, 
    							ControlSpec(*(varCtr[i+1][j].copyFromStart(3))), 
      	  						{|ez| 
      	  							environment.at(varCtr[i]).put(j, ez.value); 
      	  				 			(saveData[0][i+1][j]).put(4, ez.value); 
      	  						}, 
      	  					varCtr[i+1][j].at(4), labelWidth: labelWidth);
      	  				ez.varGUIcolorize(tempColors[varCtrColors[k], numColorExp])
  	    	  			);
    	  	  			k = k+1;
      			})	
	   	 	}{
    				str = varCtr[i].asString ++ " ";
    				cview.decorator.nextLine;
    				varStrings = varStrings.add(str);
	
	    			varCtrSliders = varCtrSliders.add(    					ez = EZSlider.new(cview, sliderWidth @ sliderHeight, str, 
    						ControlSpec(*varCtr[i+1].copyFromStart(3)), 
      	  				{|ez| 
      	  					environment.put(varCtr[i], ez.value); 
      	  					(saveData[0][i+1]).put(4, ez.value);  
      	  				}, 
      	  				varCtr[i+1].at(4), labelWidth: labelWidth);
   		   	  		ez.varGUIcolorize(tempColors[varCtrColors[k], numColorExp]);
   		   		);
      		k = k+1;	
			};	
    		});

		(synthCtrNum != 0).if {
			cview.decorator.reset;
			secondColumnIndent = (varCtrNum != 0).if { varCtrNum.sign * sliderWidth + colMargin }{ 0 };
			cview.decorator.shift(secondColumnIndent);
		};
		
		k = 0;
		forBy(0, synthCtr.size - 2, 2, {|i|
    			forBy(0, synthCtr[i+1].size - 2, 2, {|j|
    				var str, synth, ctrStr;
		   		cview.decorator.nextLine;
				cview.decorator.shift(secondColumnIndent +
					(((j == 0).if { 0 } { 1 }) * (gap + labelWidth)));
 		   		((j == 0).if { synthNameBoxes = synthNameBoxes.add(
 		   			GUI.staticText.new(cview, Rect(0, 0, labelWidth, sliderHeight)) 
 		   				.background_(tempColors[synthCtrColors[k] + varColorNum].exp(numColorExp))
 		   				.string_(" " ++ (synthCtr[i].asString))
 		   				.align_(\center)
 		   			);
				});
				synthCtr[i+1][j+1].every(_.isSequenceableCollection).if {
    					synthCtr[i+1][j+1].do({|item,l|
	    					(l > 0).if{
		   					cview.decorator.nextLine;
							cview.decorator.shift(secondColumnIndent + gap + labelWidth);
	    					};
   			 			str = synthCtr[i+1][j].asString ++ "[" ++ l.asString ++ "]" ++ " ";
						synth = synthIDs[i/2];
						ctrStr = synthCtr[i+1][j].asString;
    						synthCtrStrings = synthCtrStrings.add(str);
  		   				synthCtrSliders = synthCtrSliders.add(    							EZSlider(cview, sliderWidth @ sliderHeight, str, 
    							ControlSpec(*synthCtr[i+1][j+1][l].copyFromStart(3)), 
      	  						{|ez| 
	      	  						server.listSendMsg(this.makeMsgList(synth, ctrStr, l, ez.value, i+1, j+1));
      	  							(saveData[1][i+1][j+1][l]).put(4, ez.value);
      	  						}, synthCtr[i+1][j+1][l].at(4), labelWidth: labelWidth)
	      	  					.varGUIcolorize(tempColors[synthCtrColors[k] + varColorNum], numColorExp));
      					k = k+1;
   						
    					})	
					
				}{
					str = synthCtr[i+1][j].asString ++ " ";
					synth = synthIDs[i/2];
					synth = synth.isKindOf(Synth).if { synth.nodeID } { synth };
					ctrStr = synthCtr[i+1][j].asString;
    					synthCtrStrings = synthCtrStrings.add(str);
 		   			synthCtrSliders = synthCtrSliders.add(    						EZSlider(cview, sliderWidth @ sliderHeight, str, 
    						ControlSpec(*synthCtr[i+1][j+1].copyFromStart(3)), 
      	  					{|ez| 
      	  						server.sendMsg("/n_set", synth, ctrStr, ez.value);
      	  						(saveData[1][i+1][j+1]).put(4, ez.value);
      	  					}, synthCtr[i+1][j+1].at(4), labelWidth: labelWidth)
	      	  				.varGUIcolorize(tempColors[synthCtrColors[k] + varColorNum], numColorExp));
      				k = k+1;
				}
 	   		});	
    		});

		cview.decorator.reset;
		buttonRow = ((varCtrNum == 0).if { synthCtrNum }  { max(varCtrNum, synthCtrNum) }) + 1;
		buttonRowShift = (buttonRow + 1) * sliderHeight + ((buttonRow + 2) * gap);
		cview.decorator.shift(y: buttonRowShift);  
				
		updateButton = Button(cview, labelWidth @ (sliderHeight * buttonHeightFactor));
		updateButton.states = [["update", Color.black, Color.new(buttonColorHi, buttonColorHi, buttonColorLo)]];
		updateButton.action = { this.update };

		cview.decorator.shift(buttonGap);
    		buttonIndex = varCtrNum + synthCtrNum + synthCtrIDs.size;

		saveButton = Button(cview, labelWidth @ (sliderHeight * buttonHeightFactor));
    		saveButton.states = [["save as", Color.black, Color.new(buttonColorLo, buttonColorHi, buttonColorLo)]];
    		saveButton.bounds = Rect(leftMargin + labelWidth + buttonGap, upperMargin + buttonRowShift, 
    		labelWidth, sliderHeight * buttonHeightFactor);
 
		saveButton.action = { 
			Dialog.savePanel({ arg path;
				saveData.asCompileString.write(path);
			},{ });
		};
    		
    		cview.decorator.shift(buttonGap);
    		heightK = (maxFontHeight - minFontHeight) / (maxFontSliderHeight - minFontSliderHeight);
    		fontHeight = fontHeight ?? case
    		{ sliderHeight < minFontSliderHeight }{ minFontHeight }    			{ sliderHeight > maxFontSliderHeight }{ maxFontHeight }
    		{ true } { (sliderHeight - minFontSliderHeight) * heightK + minFontHeight };	
		stopButton = Button(cview, labelWidth @ sliderHeight);
		stopButton.states = [["stop", Color.black, Color.new(buttonColorHi, buttonColorLo, buttonColorLo)]];
 	  	stopButton.bounds = Rect((labelWidth + buttonGap) * 2 + leftMargin, 
    		upperMargin + buttonRowShift, labelWidth, sliderHeight * buttonHeightFactor);
		
		stopButton.action = { 
			players.do({|item| item.free; item.stop; });
			synthIDs.do({|item| server.sendMsg("/n_free", item.isKindOf(Synth).if { item.nodeID } { item })});
		};
    		this.font(font, fontHeight, buttonFontFactor);
    		w.front;
	}

	update {
		varCtrSliders.do({|item| item.action.value(item) });
		synthCtrSliders.do({|item| item.action.value(item) });
	}
	
	font {|font = "Arial", height = 16, buttonFontFactor = 1|
		varCtrSliders.do({|item| 
			item.labelView.font = Font(font, height);
			item.numberView.font = Font(font, height);
		});
		synthCtrSliders.do({|item| 
			item.labelView.font = Font(font, height);
			item.numberView.font = Font(font, height);
		});
		synthNameBoxes.do({|item|
			item.font = Font(font, height);
		});
		[updateButton, saveButton, stopButton].do({|item|
			item.font = Font(font, height * buttonFontFactor);
		});		
	}
	
	initVarCtrGroups {|varCtrGroups| // default variable color grouping
		var j = 0, col = [], subcol = [];
		^(varCtrGroups ?? (varCtr.do({|x,i|
			i.odd.if {
				x[0].isSequenceableCollection.if {
					x.do({ 
						subcol = subcol.add(j); 
						j = j + 1; 
					});
					col = col.add(subcol);
					subcol = []
				}{
					j = j + 1;
				}
			}
		});
		col))
	}
	
	initSynthCtrGroups {|synthCtrGroups| // default controller color grouping
		var j = 0, col = [], subcol = [];
		^(synthCtrGroups ?? (synthCtr.do({|x,i|
			i.odd.if {
				x.do({|y,k|
					k.odd.if {
						y[0].isSequenceableCollection.if {
							y.do({
								subcol = subcol.add(j);
								j = j + 1;
							});
						}{
							subcol = subcol.add(j);
							j = j + 1;
						}
					}
				});
				col = col.add(subcol);
				subcol = [];
			}
		});
		col))
	}

	makeMsgList {|synth, argName, index, value, saveDataI, saveDataJ|
		// synth may be Synth or nodeID
		var ctrIndex, msgList, synthDesc;
		(index > 0).if {
			(synth.isKindOf(Synth) 
				and: { (synthDesc = SynthDescLib.global.at(synth.defName.asSymbol)).notNil }
				and: { (synth.defName[0..4] == "temp_";).not } 
				and: { (ctrIndex = synthDesc.controls.collect({|x| x.name.asSymbol }).indexOf(argName.asSymbol)).notNil }).if {
					^[\n_set, synth.nodeID, ctrIndex + index, value];
				}{
					^[\n_setn, synth.isKindOf(Synth).if { synth.nodeID } { synth }, argName, index+1] ++ 
						(saveData[1][saveDataI][saveDataJ][0..index-1]).collect(_.at(4)) ++ [value];    
				}
			}{
				^[\n_set, synth.isKindOf(Synth).if { synth.nodeID } { synth }, argName, value];
		}
	}

	updateVarSliders {|key, val, indexOffset = 0, updateVar = true|
		// val may be sequenceable collection
		var startIndex = this.varCtrIDs.findKeyForValueLists(key);
		val.isSequenceableCollection.if {
			val.do {|x,i| varCtrSliders[startIndex + indexOffset + i].value = x; }
		}{
			varCtrSliders[startIndex + indexOffset].value = val;
		};
	^updateVar.if { this.update } { this }
	}

	updateSynthSliders {|synthKey, key, val, indexOffset = 0, updateSynth = true|
		// synthKey may be given synth label (symbol) or number 
		// val may be sequenceable collection
		var sKey, startIndex;
		sKey = synthKey.isInteger.if { saveData[1][synthKey * 2] }{ synthKey };
		startIndex = this.synthCtrIDs.findKeyForValueLists1(sKey, key);
		val.isSequenceableCollection.if {
			val.do {|x,i| synthCtrSliders[startIndex + indexOffset + i].value = x; }
		}{
			synthCtrSliders[startIndex + indexOffset].value = val;
		};
	^updateSynth.if { this.update } { this }
	}
			
}

